@@ -4,8 +4,8 @@ module Bundler
   # Represents metadata from when the Bundler gem was built.
   module BuildMetadata
     # begin ivars
-    @built_at = "2018-10-05".freeze
-    @git_commit_sha = "f66c33467".freeze
+    @built_at = "2018-05-16".freeze
+    @git_commit_sha = "9f7bf0ac3".freeze
     @release = true
     # end ivars
 
@@ -25,15 +25,7 @@ module Bundler
 
     # The SHA for the git commit the bundler gem was built from.
     def self.git_commit_sha
-      return @git_commit_sha if @git_commit_sha
-
-      # If Bundler has been installed without its .git directory and without a
-      # commit instance variable then we can't determine its commits SHA.
-      git_dir = File.join(File.expand_path("../../..", __FILE__), ".git")
-      return "unknown" unless File.directory?(git_dir)
-
-      # Otherwise shell out to git.
-      @git_commit_sha = Dir.chdir(File.expand_path("..", __FILE__)) do
+      @git_commit_sha ||= Dir.chdir(File.expand_path("..", __FILE__)) do
         `git rev-parse --short HEAD`.strip.freeze
       end
     end
